unit arena2_vid;
interface
 uses arena2_gfx;

 type
  TVIDHeader = packed record
   ID: array[0..2] of char;
   Unknown1: word;
   FrameCount: word;
   FrameWidth: word;
   FrameHeight: word;
   GlobalDelay: word;
   Unknown2: word;
  end;

  TVIDBlock = packed record
   case BlockType: byte of // Mildly annoying, but TYPE is a reserved word.
    // Video (headers only)
    1, 3, 4: (Delay: word);
    4: (YOffset: word);

    // Palette
    2: (Colors: TPAL);

    // Audio (headers only)
    $7C: (
     Unknown1: word;
     PlaybackRate: byte;
    );
    $7C, $7D: (DataLength: word);
   // This is the CASE that never ENDs; yes, it goes on and on my friends.... =P
  end;


 const
  VIDBLOCK_V01 : byte = $01; // DiffRLE'd video frame
  VIDBLOCK_V03 : byte = $03; // Full RLE'd frame (usually first)
  VIDBLOCK_V04 : byte = $04; // DiffRLE'd video frame with Y offset
  VIDBLOCK_PAL : byte = $02; // VGA (0..63) palette
  VIDBLOCK_AU1 : byte = $7C; // First block of audio
  VIDBLOCK_AUD : byte = $7D; // Audio block
  VIDBLOCK_EOF : byte = $14; // EOF marker

 { Places a compressed version of the given frame in OutBuf, and returns the number of bytes used. }
 function VidCompressFrame (PrevFrame: PByte; ThisFrame: PByte; const FrameDelay: word; const FrameWidth: word; const FrameSize: longword; OutBuf: PByte; const OutBufSize: longword): longword;

implementation

//******************************************************************************

 function VidCompressFrame (
  PrevFrame: PByte; // Pointer to previous frame (if any)
  ThisFrame: PByte; // Pointer to current frame
  const FrameDelay: word; // (FrameDelay + GlobalDelay) * 185 = delay in audio samples = size of preceding audio block (if any)
  const FrameWidth: word; // Width of frame(s)
  const FrameSize: longword; // Size of frame(s) in bytes
  OutBuf: PByte; // Where to put compressed data
  const OutBufSize: longword // Size of output buffer (ideally (FrameSize + (FrameSize / 127) + 2) bytes)
  // A callback function which could flush the output buffer would be nice
 ): longword;

 var
  RunLength, offX, offZ: longword;
  WalkLength, PrevByte: byte;
  Y: word;

 begin
  offX := 0; RunLength := 0; WalkLength := 0;
  Move(FrameDelay, OutBuf[1], SizeOf(FrameDelay)); // OutBuf[1] := FrameDelay;

  offZ := 3;
  if (PrevFrame = nil) then //*********** Create 0x03 block **********
   begin
    OutBuf[0] := $03;

    PrevByte := not ThisFrame[0]; // =P
    while (offX < FrameSize) do
     begin
      inc(WalkLength, 1);
      if (ThisFrame[offX] = PrevByte) then //***** Run *****
       begin
        inc(RunLength, 1);

        if (RunLength = 127) then // RLE string has reached max size.
         begin
          OutBuf[offZ] := $FF;
          OutBuf[offZ + 1] := PrevByte;
          inc(offZ, 2);
          RunLength := 0;
          WalkLength := 0;
         end
        else if (RunLength = 3) and (WalkLength > 3) then // End of literal string.
         begin
          OutBuf[offZ] := (WalkLength - 3);
          Move(ThisFrame[(offX + 1) - WalkLength], OutBuf[offZ + 1], (WalkLength - 3));
          inc(offZ, (WalkLength - 2));
          WalkLength := 3;
         end
        else if (WalkLength = 127) then // Literal string has reached max size.
         begin
          OutBuf[offZ] := $7F;
          Move(ThisFrame[(offX + 1) - WalkLength], OutBuf[offZ + 1], WalkLength);
          inc(offZ, 128);
          WalkLength := 0;
          RunLength := 0;
         end;

       end
      else //*************** Walk (don't run) **************
       begin
        if (RunLength > 2) then // End of RLE string.
         begin
          WalkLength := (RunLength + 128);
          OutBuf[offZ] := WalkLength;
          OutBuf[offZ + 1] := (PrevByte);
          WalkLength := 1;
          inc(offZ, 2);
         end
        else if (WalkLength = 127) then // Literal string has reached max size.
         begin
          OutBuf[offZ] := $7F;
          Move(ThisFrame[(offX + 1) - WalkLength], OutBuf[offZ + 1], WalkLength);
          inc(offZ, 128);
          WalkLength := 0;
         end;

         PrevByte := ThisFrame[offX];
         RunLength := 1;
       end;

      inc(offX, 1);
     end;

    if (RunLength > 2) then
     begin
      WalkLength := (RunLength + 128);
      OutBuf[offZ] := WalkLength;
      OutBuf[offZ + 1] := (PrevByte);
      inc(offZ, 2);
     end
    else if (WalkLength > 0) then
     begin
      OutBuf[offZ] := WalkLength;
      Move(ThisFrame[(offX + 1) - WalkLength], OutBuf[offZ + 1], WalkLength);
      inc(offZ, (WalkLength + 1));
     end;

   end
//********************************************************************
  else //***************** Create 0x01 or 0x04 block *****************
   begin
    OutBuf[0] := $01;

    while (offX < FrameSize) do
     begin
      if (ThisFrame[offX] = PrevFrame[offX]) then
       begin
        inc(RunLength, 1);

        if (RunLength = 2) and (WalkLength > 1) then // End of literal string.
         begin
          OutBuf[offZ] := (WalkLength - 1);
          Move(ThisFrame[offX - WalkLength], OutBuf[offZ + 1], (WalkLength - 1));
          inc(offZ, WalkLength); // ?
          WalkLength := 0;
         end
        else if (RunLength < 2) then
         begin
          inc(WalkLength, 1);
          if (WalkLength = 127) then // Literal string has reached max size.
           begin
            OutBuf[offZ] := WalkLength;
            Move(ThisFrame[(offX + 1) - WalkLength], OutBuf[offZ + 1], WalkLength);
            inc(offZ, 128);
            WalkLength := 0;
            RunLength := 0; // ?
           end;
         end;

       end
      else
       begin
        inc(WalkLength, 1);

        if (offZ = 3) and (RunLength > FrameWidth) then
         begin
          OutBuf[0] := $04;
          Y := RunLength div FrameWidth;
          Move(Y, OutBuf[3], SizeOf(Y));
          RunLength := RunLength mod FrameWidth;
          offZ := 5;
         end;
        if (RunLength > 1) then // End of RLE string.
         begin
          if (RunLength > 127) then
           begin
            FillByte(OutBuf[offZ], (RunLength div 127), $FF);
            inc(offZ, (RunLength div 127));
            RunLength := RunLength mod 127;
            if (RunLength > 0) then
             begin
              OutBuf[offZ] := (RunLength + 128);
              inc(offZ, 1);
             end;
           end
          else
           begin
            OutBuf[offZ] := (RunLength + 128);
            inc(offZ, 1);
           end;
          WalkLength := 1;
         end
        else if (WalkLength = 127) then // Literal string has reached max size.
         begin
          OutBuf[offZ] := WalkLength;
          Move(ThisFrame[(offX + 1) - WalkLength], OutBuf[offZ + 1], (WalkLength));
          inc(offZ, 128);
          WalkLength := 0;
         end;

         RunLength := 0;
       end;

      inc(offX, 1);
     end;

    if (WalkLength > 0) then
     begin
      OutBuf[offZ] := WalkLength;
      Move(ThisFrame[(offX + 1) - WalkLength], OutBuf[offZ + 1], WalkLength);
      inc(offZ, (WalkLength + 1));
     end;
    OutBuf[offZ] := $00;
    inc(offZ, 1);
   end;

  exit(offZ);
 end; // VidCompressFrame

end.
