{$inline on}
program vidtitle;
uses
  sysutils, strutils,
  arena2_vid, arena2_gfx;

const
  RawTagDelims = ['<', '|', '>'];

type
  TRGBQuad = packed record
    B: byte;
    G: byte;
    R: byte;
    O: byte; // Always 0
  end;
  TBMPHeader = packed record
    // BITMAPFILEHEADER
    ID: packed array[1..2] of char; // 'BM'
    FileSize: longword; // DataSize + Offset
    Res1: word;
    Res2: word;
    Offset: longword; // 1078 = SizeOf(TBMPHeader) + SizeOf(Palette)
    // BITMAPINFO <CORE>
    IHSize: longword; // 40
    Width: longword;
    Height: longword;
    Planes: word; // 1
    BitCount: word; // 8
    // </CORE>
    Compress: longword;
    DataSize: longword; // Width * Height
    XPxPerMeter: longword; // dpi * 39.37ish
    YPxPerMeter: longword; // dpi * 39.37ish
    Colors: longword; // 256
    IColors: longword; // 256
  end;
  TBGR = packed record
    B: byte;
    G: byte;
    R: byte;
  end;
  Sextet = 0..63;
  TYeOldeRGB = packed record
    R: Sextet;
    G: Sextet;
    B: Sextet;
  end;

var
  T: longword = 0;
  Header: TVIDHeader;
  Palette: TPAL;
  PlaybackRate: byte;
  FrameSize: longword = 64000; // default - 320x200
  ThisFrame, PrevFrame, ModifiedFrame: packed array[0..65535] of byte;
  OverlayData: packed array[0..65535] of byte;
  MaskData: packed array[0..65535] of ByteBool;
  CompressedFrame: packed array[0..66100] of byte;
  CompressedFrameSize, TargetTime: longword;
  LightestColor: byte = 255;
  DarkestColor: byte = 0;

//****************************************************************************//
//****************************** Misc Functions ******************************//
//****************************************************************************//

function EuclideanDistance3D (
  X1: longint; Y1: longint; Z1: longint;
  X2: longint; Y2: longint; Z2: longint
) : real;
begin
  EuclideanDistance3D := sqrt(sqr(abs(X1 - X2)) + sqr(abs(Y1 - Y2)) + sqr(abs(Z1 - Z2)));
end;

function SquaredEuclideanDistance3D (
  X1: longint; Y1: longint; Z1: longint;
  X2: longint; Y2: longint; Z2: longint
) : longword;
begin
  SquaredEuclideanDistance3D := sqr(abs(X1 - X2)) + sqr(abs(Y1 - Y2)) + sqr(abs(Z1 - Z2));
end;

//********************************************************//
function FastPercievedColorDistance (
  R1: longint; G1: longint; B1: longint;
  R2: longint; G2: longint; B2: longint
) : longword;
begin
  FastPercievedColorDistance := sqr(30 * abs(R1 - R2)) + sqr(59 * abs(G1 - G2)) + sqr(11 * abs(B1 - B2));
end;

//********************************************************//
function SixToEight (b: Sextet) : byte; inline;
begin
  SixToEight := (b shl 2) or (b shr 4);
end;

//********************************************************//
function ProcessVIDBlocks (var InputFile: file; var OutputFile: file; TargetTime: longword): boolean;
var
  b, BlockType, RunLength: byte;
  AudioData: packed array[0..65535] of byte; 
  i, YOffset, FrameDelay, AudioLength: word;
  Z: longword;

begin
  repeat
    writeln(FilePos(InputFile), '->', FilePos(OutputFile));
    BlockRead(InputFile, BlockType, SizeOf(BlockType)); // get block type
    write(BlockType, ': ');
//************************************//
    if (BlockType = VIDBLOCK_PAL) then
     begin
      BlockWrite(OutputFile, BlockType, SizeOf(BlockType));
      writeln('Palette');
      BlockRead(InputFile, Palette, SizeOf(Palette));
      BlockWrite(OutputFile, Palette, SizeOf(Palette));
      for i := 0 to 255 do
       begin
        Z := (Palette[i].R * 30) + (Palette[i].G * 59) + (Palette[i].B * 11);
        if (Z > ((Palette[LightestColor].R * 30) + (Palette[LightestColor].G * 59) + (Palette[LightestColor].B * 11))) then
          LightestColor := i;

        if (Z < ((Palette[DarkestColor].R * 30) + (Palette[DarkestColor].G * 59) + (Palette[DarkestColor].B * 11))) then
          DarkestColor := i;
       end;
     end

//************************************//
    else if (BlockType in [VIDBLOCK_V03, VIDBLOCK_V01, VIDBLOCK_V04]) then
     begin
      BlockRead(InputFile, FrameDelay, SizeOf(FrameDelay));
      writeln('Video frame; T = ', T);

      if (BlockType = VIDBLOCK_V04) then
       begin
        BlockRead(InputFile, YOffset, SizeOf(YOffset));
        Z := (YOffset * Header.FrameWidth);
       end
      else
        Z := 0;

      // decode frame
      repeat
        BlockRead(InputFile, RunLength, SizeOf(RunLength));
        if (RunLength > 127) then
         begin
          if (BlockType = VIDBLOCK_V03) then
           begin
            BlockRead(InputFile, b, SizeOf(b));
            FillByte(ThisFrame[Z], (RunLength - 128), b);
           end;

          inc(Z, (RunLength - 128));
         end
        else
         begin
          BlockRead(InputFile, ThisFrame[Z], RunLength);
          inc(Z, RunLength);
         end;
      until (RunLength = 0) or (Z >= FrameSize);

      // apply overlay
      for i := Low(OverlayData) to High(OverlayData) do
        if MaskData[i] then
          ModifiedFrame[i] := OverlayData[i]
        else
          ModifiedFrame[i] := ThisFrame[i];

      if (BlockType = VIDBLOCK_V03) then
        CompressedFrameSize := VidCompressFrame(nil, @ModifiedFrame, FrameDelay, Header.FrameWidth, FrameSize, @CompressedFrame, SizeOf(CompressedFrame))
      else
        CompressedFrameSize := VidCompressFrame(@PrevFrame, @ModifiedFrame, FrameDelay, Header.FrameWidth, FrameSize, @CompressedFrame, SizeOf(CompressedFrame));
      BlockWrite(OutputFile, CompressedFrame, CompressedFrameSize);

      PrevFrame := ModifiedFrame;
      inc(T, (FrameDelay + Header.GlobalDelay));

      if (T >= TargetTime) then
        exit(true);
     end

//************************************//
    else if (BlockType = VIDBLOCK_AU1) then
     begin
      BlockWrite(OutputFile, BlockType, SizeOf(BlockType));

      BlockRead(InputFile, i, SizeOf(i));
      BlockWrite(OutputFile, i, SizeOf(i));

      BlockRead(InputFile, PlaybackRate, SizeOf(PlaybackRate));
      BlockWrite(OutputFile, PlaybackRate, SizeOf(PlaybackRate));

      BlockRead(InputFile, AudioLength, SizeOf(AudioLength));
      BlockWrite(OutputFile, AudioLength, SizeOf(AudioLength));

      writeln('Audio, ', (1000000 DIV (256 - PlaybackRate)), 'Hz, ', AudioLength, ' samples');

      BlockRead(InputFile, AudioData, AudioLength);
      BlockWrite(OutputFile, AudioData, AudioLength);
     end

//************************************//
    else if (BlockType = VIDBLOCK_AUD) then
     begin
      BlockWrite(OutputFile, BlockType, SizeOf(BlockType));

      BlockRead(InputFile, AudioLength, SizeOf(AudioLength));
      BlockWrite(OutputFile, AudioLength, SizeOf(AudioLength));

      writeln('Audio,  ', AudioLength, ' samples');

      BlockRead(InputFile, AudioData, AudioLength);
      BlockWrite(OutputFile, AudioData, AudioLength);
     end

//************************************//
    else if (BlockType = 0) then
      writeln('Disregard this')

    else
     begin
      BlockWrite(OutputFile, BlockType, SizeOf(BlockType));
      if (BlockType = VIDBLOCK_EOF) then
        writeln('EOF')
      else
        writeln('???');
     end;

  until eof(InputFile);

  exit(false);
end;

//****************************************************************************//
//*********************************** Main ***********************************//
//****************************************************************************//
var
  RawFile: text;
  RawLine, RawTag, RawToken: AnsiString;
  BinRow: bitpacked array[0..511] of boolean;
  HexRow: bitpacked array[0..511] of 0..15;
  BMPRow: packed array[0..511] of TBGR;
  buf: packed array[0..511] of byte;
  BMPPalette: array[0..255] of TRGBQuad;
  ColorDist, ClosestColorDist: longword;
  CLUT: array[0..255] of byte;
  VidFN, BMPFN, OutputFN, TempStr: string;
  InputFile, OutputFile, BMPFile: file;
  BMPHeader: TBMPHeader;
  HaveVID: boolean = false;
  ForceRemap: boolean;
  i, j: word;
  X, Y, Z, N: longword;

begin
  if (ParamCount < 1) then
   begin
    writeln('Usage: ' + ParamStr(0) + ' <Input File>');
    halt(1)
   end
  else if not FileExists(ParamStr(1)) then
   begin
    writeln(ParamStr(1), ': File not found.');
    halt(1)
   end;
  FillByte(MaskData, SizeOf(MaskData), 0);
  assign(RawFile, ParamStr(1));
  reset(RawFile);

  while not eof(RawFile) do
   begin
    readln(RawFile, RawLine);
    RawLine := Trim(RawLine);

    while (pos('<', RawLine) > 0) do
     begin
      X := pos('<', RawLine);
      RawTag := copy(RawLine, X, Length(RawLine));
      SetLength(RawTag, pos('>', RawTag));
      Y := X + Length(RawTag);
      RawLine := copy(RawLine, Y, Length(RawLine));

      if (WordCount(RawTag, RawTagDelims) > 0) then
       begin
        writeln(RawTag);
        RawToken := upcase(ExtractWord(1, RawTag, RawTagDelims));

	//************************************//
        if (RawToken = 'CLEAR') then
          FillByte(MaskData, SizeOf(MaskData), 0)

	//************************************//
        else if HaveVID and (RawToken = 'NEXT') then
         begin // apply overlay to current frame, then continue reading tokens
          HaveVID := ProcessVIDBlocks(InputFile, OutputFile, T);
          if not HaveVID then
           begin // finish current vid
            close(InputFile);
            close(OutputFile);
           end;
         end

	//************************************//
        else if HaveVID and (RawToken = 'KEEP') then
         begin
          ThisFrame := ModifiedFrame;

{$IFDEF BUGHUNT} // alignment check
          ThisFrame[1] := LightestColor;
          ThisFrame[2] := DarkestColor;
          ThisFrame[3] := LightestColor;
{$ENDIF}
         end

	//************************************//
        else if HaveVID and (RawToken = 'PROCEED_TO_END') then
         begin // continue to process VID blocks until EOF is reached
          while ProcessVIDBlocks(InputFile, OutputFile, $ffffffff) do
            continue;
          close(InputFile);
          close(OutputFile);
          HaveVID := false;
         end

	//**************************//
        else if (WordCount(RawTag, RawTagDelims) > 1) then
         begin
	//************************************//
          if (RawToken = 'VID') then
           begin // get vid filename
            VidFN := ExtractWord(2, RawTag, RawTagDelims);
            if (WordCount(RawTag, RawTagDelims) > 2) then
             begin
              OutputFN := ExtractWord(3, RawTag, RawTagDelims);
              if FileExists(OutputFN) then
               begin
                TempStr := OutputFN + '.bak';
                Z := 1;
                while FileExists(TempStr) do
                 begin
                  inc(Z, 1);
                  WriteStr(TempStr, OutputFN, Z, '.bak');
                 end;

                if RenameFile(OutputFN, TempStr) then
                 begin
                  writeln('Backed up original ', OutputFN, ' to ', TempStr);
                  if (ExpandFileName(OutputFN) = ExpandFileName(VidFN)) then
                    VidFN := TempStr;
                 end
                else
                  OutputFN := VidFN + 'Titled.VID';
               end;
             end
            else
              OutputFN := VidFN + 'Titled.VID';

            if HaveVID then
             begin // finish current vid
              while ProcessVIDBlocks(InputFile, OutputFile, $ffffffff) do
                continue;

              close(InputFile);
              close(OutputFile);
             end;

            if not FileExists(VidFN) then
             begin
              HaveVID := false;
              continue; // next token
             end;

            Filemode := 0; // should be Read - default = ReadWrite = run-time error when opening read-only files
            assign(InputFile, VidFN);
            reset(InputFile, 1);
            Filemode := 2; // back to ReadWrite - just in case leaving it as Read does something even stupider

            BlockRead(InputFile, Header, SizeOf(Header));
            writeln(Header.ID, ' header:');
            writeln('  FrameCount: ', Header.FrameCount);
            writeln('  FrameWidth: ', Header.FrameWidth);
            writeln('  FrameHeight: ', Header.FrameHeight);
            writeln('  GlobalDelay: ', Header.GlobalDelay);
            writeln('  Unknown2: ', Header.Unknown2);

            FrameSize := (Header.FrameWidth * Header.FrameHeight);

            FillByte(ThisFrame, SizeOf(ThisFrame), 0);
            FillByte(MaskData, SizeOf(MaskData), 0);

            assign(OutputFile, OutputFN);
            rewrite(OutputFile, 1);
            BlockWrite(OutputFile, Header, SizeOf(Header));
            T := 0;
            HaveVID := true;
           end

	//************************************//
          else if (RawToken = 'OVERLAY') then
           begin // load overlay data
            BMPFN := ExtractWord(2, RawTag, RawTagDelims);
            if FileExists(BMPFN) then
             begin
              assign(BMPFile, BMPFN);
              reset(BMPFile, 1);

              BlockRead(BMPFile, BMPHeader, SizeOf(BMPHeader));

              if (BMPHeader.ID <> 'BM') then
                writeln(BMPFN, ' is not a valid BMP file!')

              else if (BMPHeader.Width <> Header.FrameWidth) or (BMPHeader.Height <> Header.FrameHeight) then
               begin
                writeln(BMPFN, ' has the wrong dimensions!');
                writeln('Correct dimensions for current VID are: ', Header.FrameWidth, 'x', Header.FrameHeight);
               end

              else if (BMPHeader.Compress <> 0) then
               begin
                writeln(BMPFN, ' is compressed!');
                writeln('Could not load image data.');
               end

              else
               begin
                ForceRemap := (WordCount(RawTag, RawTagDelims) > 2) and (upcase(ExtractWord(3, RawTag, RawTagDelims)) = 'REMAP');

                if (BMPHeader.IHSize > 40) then // fucking ImageMagick
                  seek(BMPFile, (FilePos(BMPFile) + BMPHeader.IHSize - 40));

                if (BMPHeader.BitCount = 8) then
                 begin
                  BlockRead(BMPFile, BMPPalette, SizeOf(BMPPalette));
                  if ForceRemap then
                   begin
                    for i := 0 to 255 do
                     begin // populate Color Look-Up Table
                      ClosestColorDist := $ffffffff;
                      for j := 0 to 255 do
                       begin
                        ColorDist := FastPercievedColorDistance(BMPPalette[i].R, BMPPalette[i].G, BMPPalette[i].B, (Palette[j].R * 4), (Palette[j].G * 4), (Palette[j].B * 4));
                        if (ColorDist < ClosestColorDist) then
                         begin
                          ClosestColorDist := ColorDist;
                          CLUT[i] := j;
                         end;
                        if (ClosestColorDist = 0) then
                          break;
                       end;
                      writeln('Mapped BMP color ', i, ' (#', HexStr(BMPPalette[i].R, 2), HexStr(BMPPalette[i].G, 2), HexStr(BMPPalette[i].B, 2), ') to VID color ', CLUT[i], ' (#', HexStr((Palette[CLUT[i]].R * 4), 2), HexStr((Palette[CLUT[i]].G * 4), 2), HexStr((Palette[CLUT[i]].B * 4), 2), ').');
                     end;
                    seek(BMPFile, BMPHeader.Offset);
                    for Y := (Header.FrameHeight - 1) downto 0 do
                     begin
                      Z := (Y * Header.FrameWidth);
{$IFDEF BUGHUNT}
                      writeln('Row ', Y, ': ', FilePos(BMPFile), ' -> ', Z);
{$ENDIF}
                      BlockRead(BMPFile, buf, (BMPHeader.DataSize DIV BMPHeader.Height));
                      for X := 0 to (Header.FrameWidth - 1) do
                        OverlayData[(Z + X)] := CLUT[buf[X]];
                     end;
                   end // ForceRemap

                  else // check BMP palette against VID palette
                   begin
                    for i := 0 to 255 do
                     begin
                      ColorDist := SquaredEuclideanDistance3D((Palette[i].R * 4), (Palette[i].G * 4), (Palette[i].B * 4), BMPPalette[i].R, BMPPalette[i].G, BMPPalette[i].B);
                      if (ColorDist > 40) then
                        writeln('WARNING: BMP palette index ', i, ' is off by ', sqrt(ColorDist):6:3, ' from VID palette.');
                     end;
                    seek(BMPFile, BMPHeader.Offset);
                    for Y := (Header.FrameHeight - 1) downto 0 do
                     begin
                      Z := (Y * Header.FrameWidth);
{$IFDEF BUGHUNT}
                      writeln('Row ', Y, ': ', FilePos(BMPFile), ' -> ', Z);
{$ENDIF}
                      BlockRead(BMPFile, buf, (BMPHeader.DataSize DIV BMPHeader.Height));
                      Move(buf, OverlayData[Z], Header.FrameWidth);
                     end;
                   end;
                 end // 8-bit BMP

                else if (BMPHeader.BitCount = 4) then
                 begin
                  BlockRead(BMPFile, BMPPalette, 64); // 16 x RGBQuad
                  for i := 0 to 15 do
                   begin // populate Color Look-Up Table
                    ClosestColorDist := $ffffffff;
                    for j := 0 to 255 do
                     begin
                      ColorDist := FastPercievedColorDistance(BMPPalette[i].R, BMPPalette[i].G, BMPPalette[i].B, (Palette[j].R * 4), (Palette[j].G * 4), (Palette[j].B * 4));
                      if (ColorDist < ClosestColorDist) then
                       begin
                        ClosestColorDist := ColorDist;
                        CLUT[i] := j;
                       end;
                      if (ClosestColorDist = 0) then
                        break;
                     end;
                    writeln('Mapped BMP color ', i, ' (#', HexStr(BMPPalette[i].R, 2), HexStr(BMPPalette[i].G, 2), HexStr(BMPPalette[i].B, 2), ') to VID color ', CLUT[i], ' (#', HexStr((Palette[CLUT[i]].R * 4), 2), HexStr((Palette[CLUT[i]].G * 4), 2), HexStr((Palette[CLUT[i]].B * 4), 2), ').');
                   end;
                  seek(BMPFile, BMPHeader.Offset);
                  for Y := (Header.FrameHeight - 1) downto 0 do
                   begin
                    Z := (Y * Header.FrameWidth);
{$IFDEF BUGHUNT}
                    writeln('Row ', Y, ': ', FilePos(BMPFile), ' -> ', Z);
{$ENDIF}
                    BlockRead(BMPFile, HexRow, (BMPHeader.DataSize DIV BMPHeader.Height));
                    for X := 0 to (Header.FrameWidth - 1) do
                      OverlayData[(Z + X)] := CLUT[HexRow[(X xor 1)]];
                   end;
                 end // 4-bit BMP

                else if (BMPHeader.BitCount = 1) then
                 begin
                  BlockRead(BMPFile, BMPPalette, 8); // 2 x RGBQuad
                  for i := 0 to 1 do
                   begin // populate Color Look-Up Table
                    ClosestColorDist := $ffffffff;
                    for j := 0 to 255 do
                     begin
                      ColorDist := FastPercievedColorDistance(BMPPalette[i].R, BMPPalette[i].G, BMPPalette[i].B, (Palette[j].R * 4), (Palette[j].G * 4), (Palette[j].B * 4));
                      if (ColorDist < ClosestColorDist) then
                       begin
                        ClosestColorDist := ColorDist;
                        CLUT[i] := j;
                       end;
                      if (ClosestColorDist = 0) then
                        break;
                     end;
                    writeln('Mapped BMP color ', i, ' (#', HexStr(BMPPalette[i].R, 2), HexStr(BMPPalette[i].G, 2), HexStr(BMPPalette[i].B, 2), ') to VID color ', CLUT[i], ' (#', HexStr((Palette[CLUT[i]].R * 4), 2), HexStr((Palette[CLUT[i]].G * 4), 2), HexStr((Palette[CLUT[i]].B * 4), 2), ').');
                   end;
                  seek(BMPFile, BMPHeader.Offset);
                  for Y := (Header.FrameHeight - 1) downto 0 do
                   begin
                    Z := (Y * Header.FrameWidth);
{$IFDEF BUGHUNT}
                    writeln('Row ', Y, ': ', FilePos(BMPFile), ' -> ', Z);
{$ENDIF}
                    BlockRead(BMPFile, BinRow, (BMPHeader.DataSize DIV BMPHeader.Height));
                    for X := 0 to (Header.FrameWidth - 1) do
                      OverlayData[(Z + X)] := CLUT[ord(BinRow[(X xor 7)])];
                   end
                 end // 1-bit BMP

                else if (BMPHeader.BitCount = 24) then
                 begin
                  seek(BMPFile, BMPHeader.Offset);
                  for Y := (Header.FrameHeight - 1) downto 0 do
                   begin
                    Z := (Y * Header.FrameWidth);
{$IFDEF BUGHUNT}
                    writeln('Row ', Y, ': ', FilePos(BMPFile), ' -> ', Z);
{$ENDIF}
                    BlockRead(BMPFile, BMPRow, (BMPHeader.DataSize DIV BMPHeader.Height));
                    for X := 0 to (Header.FrameWidth - 1) do
                     begin
                      i := 0;
                      ClosestColorDist := $ffffffff;
                      for j := 0 to 255 do
                       begin
                        ColorDist := FastPercievedColorDistance(BMPRow[X].R, BMPRow[X].G, BMPRow[X].B, (Palette[j].R * 4), (Palette[j].G * 4), (Palette[j].B * 4));
                        if (ColorDist < ClosestColorDist) then
                         begin
                          ClosestColorDist := ColorDist;
                          i := j;
                         end;
                        if (ClosestColorDist = 0) then
                          break;
                       end;
                      OverlayData[(Z + X)] := i;
                     end;
                   end;
                 end // 24-bit BMP

                else
                 begin
                  writeln(BMPFN, ' has an unsupported color depth.');
                  writeln('The recommended color depth is 256 colors/8 bits per pixel.');
                 end;
               end;

              close(BMPFile);
             end
            else
             writeln(BMPFN, ': File not found.');
           end

	//************************************//
          else if (RawToken = 'MASK') then
           begin // load mask data
            BMPFN := ExtractWord(2, RawTag, RawTagDelims);
            if FileExists(BMPFN) then
             begin
              assign(BMPFile, BMPFN);
              reset(BMPFile, 1);

              BlockRead(BMPFile, BMPHeader, SizeOf(BMPHeader));

              if (BMPHeader.ID <> 'BM') then
                writeln(BMPFN, ' is not a valid BMP file!')

              else if (BMPHeader.Width <> Header.FrameWidth) or (BMPHeader.Height <> Header.FrameHeight) then
               begin
                writeln(BMPFN, ' has the wrong dimensions!');
                writeln('Correct dimensions for current VID are: ', Header.FrameWidth, 'x', Header.FrameHeight);
               end

              else if (BMPHeader.Compress <> 0) then
               begin
                writeln(BMPFN, ' is compressed!');
                writeln('Could not load image data.');
               end

              else
               begin
                seek(BMPFile, BMPHeader.Offset);

                if (BMPHeader.BitCount = 1) then
                  for Y := (Header.FrameHeight - 1) downto 0 do
                   begin
                    Z := (Y * Header.FrameWidth);
{$IFDEF BUGHUNT}
                    writeln('Row ', Y, ': ', FilePos(BMPFile), ' -> ', Z);
{$ENDIF}
                    BlockRead(BMPFile, BinRow, (BMPHeader.DataSize DIV BMPHeader.Height));
                    for X := 0 to (Header.FrameWidth - 1) do
                      MaskData[(Z + X)] := BinRow[(X xor 7)];
                   end

                else if (BMPHeader.BitCount = 4) then
                  for Y := (Header.FrameHeight - 1) downto 0 do
                   begin
                    Z := (Y * Header.FrameWidth);
{$IFDEF BUGHUNT}
                    writeln('Row ', Y, ': ', FilePos(BMPFile), ' -> ', Z);
{$ENDIF}
                    BlockRead(BMPFile, HexRow, (BMPHeader.DataSize DIV BMPHeader.Height));
                    for X := 0 to (Header.FrameWidth - 1) do
                      MaskData[(Z + X)] := (HexRow[(X xor 1)] > 0);
                   end

                else if (BMPHeader.BitCount = 8) then
                  for Y := (Header.FrameHeight - 1) downto 0 do
                   begin
                    Z := (Y * Header.FrameWidth);
{$IFDEF BUGHUNT}
                    writeln('Row ', Y, ': ', FilePos(BMPFile), ' -> ', Z);
{$ENDIF}
                    BlockRead(BMPFile, buf, (BMPHeader.DataSize DIV BMPHeader.Height));
                    Move(buf, MaskData[Z], Header.FrameWidth);
                   end

                else if (BMPHeader.BitCount = 24) then
                  for Y := (Header.FrameHeight - 1) downto 0 do
                   begin
                    Z := (Y * Header.FrameWidth);
{$IFDEF BUGHUNT}
                    writeln('Row ', Y, ': ', FilePos(BMPFile), ' -> ', Z);
{$ENDIF}
                    BlockRead(BMPFile, BMPRow, (BMPHeader.DataSize DIV BMPHeader.Height));
                    for X := 0 to (Header.FrameWidth - 1) do
                      MaskData[(Z + X)] := ByteBool(BMPRow[X].R or BMPRow[X].G or BMPRow[X].B);
                   end

                else
                 begin
                  writeln(BMPFN, ' has an unsupported color depth.');
                  writeln('The recommended color depth is 256 colors/8 bits per pixel.');
                 end;
               end;

              close(BMPFile);
             end
            else
             begin
              writeln(BMPFN, ': File not found.');
              FillByte(MaskData, SizeOf(MaskData), 0);
             end
           end

	//************************************//
          else if HaveVID and (RawToken = 'PROCEED_TO') then
           begin // continue to process VID blocks until target time is reached
            ReadStr(ExtractWord(2, RawTag, RawTagDelims), TargetTime);

            HaveVID := ProcessVIDBlocks(InputFile, OutputFile, TargetTime);
            if not HaveVID then
             begin // finish current vid
              close(InputFile);
              close(OutputFile);
             end;
           end

	//************************************//
          else if (RawToken = 'FILL') then
           begin
            ReadStr(ExtractWord(2, RawTag, RawTagDelims), i);
            FillByte(OverlayData, SizeOf(OverlayData), (i and $FF));
           end;

         end; // tags with at least one parameter

       end; // non-empty tags

     end; // tags in line

   end; // lines in file

  if HaveVID then
   begin // finish current vid
    while ProcessVIDBlocks(InputFile, OutputFile, $ffffffff) do
      continue;

    close(InputFile);
    close(OutputFile);
   end;

  close(RawFile);

end.
